/*--------------------------------------------------
 Andreas Brietzke, Nagareyama, J
 2016/06/11
 
 ESP12
 0: Relay Heat
 2: Relay Steam
 4: spare
 5: spare
 12: Data DS18B20
 13: Input heat
 14: Input ready
 15: GND
 16: Input pump
 ADC: spare

 Arduino extensions:
 ESP8266 scetch data upload
 WIFI101 firmware updater
 
 https://www.mikrocontroller.net/topic/387776
 Websocket lib from: https://github.com/Links2004/arduinoWebSockets/tree/master/src
 example from here: https://gist.github.com/bbx10/667e3d4f5f2c0831d00b
--------------------------------------------------*/

#include <NTPClient.h>
#include <ESP8266WiFi.h>
#include <ESP8266WiFiMulti.h>
#include <WebSocketsServer.h>
#include <Hash.h>
#include <ESP8266WebServer.h>
#include <ESP8266mDNS.h>
#include <WiFiUdp.h>
#include <ArduinoOTA.h>
#include <FS.h>
#include <base64.h>
#include <OneWire.h>
#include <DallasTemperature.h>
#include <stdlib.h>
//#include <string.h>

// #define DEBUG 1

#define WIFI_TIMEOUT 20                                    // Wait for connect * 500ms 
#define MAX_PIC_SIZE  1 * 1024                             // up to this sie will be encoded in Base64

// GPIO definitions
#define ONE_WIRE_BUS 12  // DS18B20 pin
#define OUTCOMMON 0
#define OUTSTEAM 2
#define INREADY 14
#define INHEAT 13
#define INPUMP 16

#define FIRSTYEAR       1970
#define FIRSTDAY        4                                  // 0 = Sunday

OneWire oneWire(ONE_WIRE_BUS);
DallasTemperature DS18B20(&oneWire);
MDNSResponder mdns;                                         
ESP8266WiFiMulti WiFiMulti;                                 
ESP8266WebServer webserver(80);                             // Create an instance of the webserver on Port 80
WebSocketsServer webSocket = WebSocketsServer(81);
WiFiUDP ntpUDP;
NTPClient timeClient(ntpUDP, 3600*9);                      // Japan +9h

String ssid = "xxxx";                                      // default, overwrite by config.txt if exist
String password = "xxxx";
String bu_ssid = ""; 
String bu_password = "";
String ap_name = "ESP-Test";
String ap_password = "";
String ap_ip   = "192.168.4.1";
long int CoffeeTimerValue = 300;

//struct time {
//  uint8_t  second;
//  uint8_t  minute;
//  uint8_t  hour;
//  uint8_t  day;
//  uint8_t  month;
//  uint16_t year;
//  uint8_t  wday;
//};
//time *tdate;
//const uint8_t DayOfMonth[] = { 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
//const String DayOfWeek[] = {"Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"};

bool ap_mode = false;                                      // true if AP aktive  
 
void configWifi(void);                                     // Get WLAN-Data from config.txt
void WiFiStart(void);

String GetPicture(String filename);                        // A pic smaler than MAX_PIC_SIZE will be base64 coded
String getContentType(String filename);                    // set content type in depend of file extension
enum {OFF, HEAT, STEAM, READY};
uint8_t CoffeeStat;                                        // current Coffee stat
uint8_t RelayStat;                                         // Relay stat
int Coffeetimer;                                           // current Timer value
unsigned long oldTime;                                     // old value ms Timer

const char sSTEAM[] = "steam";                             // Commands sent through Web Socket
const char sHEAT[] = "heat";                               // Commands sent through Web Socket
const char sREADY[] = "ready";                             // Commands sent through Web Socket
const char sOFF[] = "off";                                 // Commands sent through Web Socket
const char sREAD[] = "read";                               // Commands sent through Web Socket

void notFound(void);                                       // send 404 message 
void handleUnknown(void);                                  // all what found in SPIFFS without callback function
void handleUpload(void);                                   // Get file and write to SPIFFS
void handleConfig(void);
void handleCoffee(void);
void SetCoffee(uint8_t, uint8_t);
void webSocketEvent(uint8_t, WStype_t, uint8_t *, size_t);


// ###############################################################
// --------------------- setup/loop ------------------------------
// ###############################################################

void setup()
{
  // start serial
  Serial.begin(115200);                           
  Serial.println("");
  Serial.println("RESET");
  
  // Initialize file system.
  if (!SPIFFS.begin())
  {
    Serial.println("SPIFFS nicht initialisiert!");   
    while(1)                                               // SPIFFS is obsolete
    {
      yield();
    }
  }
  Serial.println("SPIFFS ok");

  configWifi();
       
  // inital connect
  WiFi.mode(WIFI_STA);
  WiFiStart();

//  timeClient.begin();
  
  // Port defaults to 8266
  // ArduinoOTA.setPort(8266);

  // Hostname defaults to esp8266-[ChipID]
  ArduinoOTA.setHostname("Coffee");

  // No authentication by default
  // ArduinoOTA.setPassword((const char *)"123");

  ArduinoOTA.onStart([]() {
    Serial.println("Start");
    digitalWrite(OUTCOMMON,1);
    digitalWrite(OUTSTEAM,1);
  });
  ArduinoOTA.onEnd([]() {
    Serial.println("\nEnd");
  });
  ArduinoOTA.onProgress([](unsigned int progress, unsigned int total) {
    Serial.printf("Progress: %u%%\r", (progress / (total / 100)));
  });
  ArduinoOTA.onError([](ota_error_t error) {
    Serial.printf("Error[%u]: ", error);
    if (error == OTA_AUTH_ERROR) Serial.println("Auth Failed");
    else if (error == OTA_BEGIN_ERROR) Serial.println("Begin Failed");
    else if (error == OTA_CONNECT_ERROR) Serial.println("Connect Failed");
    else if (error == OTA_RECEIVE_ERROR) Serial.println("Receive Failed");
    else if (error == OTA_END_ERROR) Serial.println("End Failed");
  });
  ArduinoOTA.begin();
  Serial.println("Ready");
  Serial.print("IP address: ");
  Serial.println(WiFi.localIP());


  // Start the server
  webserver.on("/", handleCoffee);
  webserver.on("/coffee.html", handleCoffee);
  webserver.on("/coffee", handleCoffee);

  webserver.on("/config.html", HTTP_GET, handleConfig);
  webserver.on("/config", HTTP_GET, handleConfig);
  webserver.on("/config.html", HTTP_POST, handleConfig, handleUpload);
  webserver.on("/config", HTTP_POST, handleConfig, handleUpload);

  webserver.onNotFound(handleUnknown);
  webserver.begin();

  Serial.println("HTTP server gestartet Port 80");
  
  webSocket.begin();
  webSocket.onEvent(webSocketEvent);

  DS18B20.requestTemperatures();                           // first reading is not correct

  pinMode(OUTCOMMON, OUTPUT);
  digitalWrite(OUTCOMMON, 1);                              // off is high!
  pinMode(OUTSTEAM, OUTPUT);
  digitalWrite(OUTSTEAM, 1);                               // off is high!
  pinMode(INHEAT, INPUT);  
  pinMode(INREADY, INPUT);  
  pinMode(INPUMP, INPUT);  
}

void loop()
{

  unsigned long actTime= millis();
  
  ArduinoOTA.handle();
  
  if (!ap_mode)
  {
    // check if WLAN is connected
    if (WiFi.status() != WL_CONNECTED)
    {
      WiFiStart();
    }
  }    
  webSocket.loop();
  webserver.handleClient();

  if (digitalRead(INPUMP)==LOW) {                          // If use the pump by manual then switch off
    CoffeeStat=OFF;
    Coffeetimer=0;
  }
  
  if (actTime - oldTime >= 1000) {                         // count each second
    oldTime=actTime;
    
    if ((RelayStat==READY) && (Coffeetimer>0)) {
      Coffeetimer--;     
      if (Coffeetimer<=0) {
        CoffeeStat=OFF;
      }
    }
//    timeClient.update();
//  gettime(timeClient.getEpochTime(), tdate);
  }

  switch (CoffeeStat) {
  case OFF:
    digitalWrite(OUTCOMMON,1);
    digitalWrite(OUTSTEAM,1);
    break;
  case HEAT:
    digitalWrite(OUTCOMMON,0);
    digitalWrite(OUTSTEAM,1);
    break;
  case STEAM:
    digitalWrite(OUTCOMMON,0);
    digitalWrite(OUTSTEAM,0);
    break;
  }
  
  RelayStat=OFF; 
  if (digitalRead(INHEAT)==LOW) {
    if (CoffeeStat==STEAM) RelayStat=STEAM;
    else if (CoffeeStat==HEAT ) RelayStat=HEAT;
  }
  if (digitalRead(INREADY)==LOW) RelayStat=READY;
  
}

// ###############################################################
// ---------------------- WiFi -----------------------------------
// ###############################################################

void configWifi()
{
#ifdef DEBUG
  Serial.println("\r\n--- configWifi()  ---");
#endif
  
  File configfile = SPIFFS.open("/config.txt","r");        // get config.txt for WLAN access
  if (configfile)
  {

    while(String indaten = configfile.readStringUntil('\n'))
    {
      if (indaten.length() == 0)
      {
        break;
      }  
      if (indaten.startsWith("user"))                      // username WLAN-Client
      {
        indaten.replace("user","");
        indaten.trim();
        ssid = indaten;  
      }
      else if (indaten.startsWith("pass"))                 // password WLAN-Client
      {
        indaten.replace("pass","");
        indaten.trim();
        password = indaten;
      }
      else if (indaten.startsWith("buser"))                // 2nd username WLAN-Client
      {
        indaten.replace("buser","");
        indaten.trim();
        bu_ssid = indaten;  
      }
      else if (indaten.startsWith("bpass"))                // 2nd password WLAN-Client
      {
        indaten.replace("bpass","");
        indaten.trim();
        bu_password = indaten;
      }
      else if (indaten.startsWith("ap_name"))              // accesspoint name WLAN-accesspoint
      {
        indaten.replace("ap_name","");
        indaten.trim();
        ap_name = indaten;
      }
      else if (indaten.startsWith("ap_pass"))              // password WLAN-accesspoint
      {
        indaten.replace("ap_pass","");
        indaten.trim();
        ap_password = indaten;
      }
      else if (indaten.startsWith("ap_ip"))                // IP WLAN-accesspoint
      {
        indaten.replace("ap_ip","");
        indaten.trim();
        ap_ip = indaten;
      }
      else if (indaten.startsWith("timer"))                // IP WLAN-accesspoint
      {
        indaten.replace("timer","");
        indaten.trim();
        CoffeeTimerValue = indaten.toInt();
      }
    }  
    configfile.close();
  }
}

// ----------------------------------------------------------------------

void WiFiStart()
{
  char ssid_send[64];
  char password_send[64];

  // Connect to WiFi network
  ssid.toCharArray(ssid_send,ssid.length() + 1);
  password.toCharArray(password_send,password.length() + 1);

  WiFi.begin(ssid_send, password_send);
#ifdef DEBUG
  Serial.print("\r\nWiFi\r\n");
  Serial.println(ssid_send);
  Serial.println(password_send);
#endif
  for (int i=0; i<WIFI_TIMEOUT; i++)
  {
    if (WiFi.status() != WL_CONNECTED)
    {
#ifdef DEBUG
      Serial.print(".");
#endif
      delay(500);
    }
  }  
  if (WiFi.status() == WL_CONNECTED)
  {
    IPAddress myAddr = WiFi.localIP();
#ifdef DEBUG
    Serial.println("connected");
    Serial.println(myAddr);
#endif
  } else {
    bu_ssid.toCharArray(ssid_send,bu_ssid.length() + 1);
    bu_password.toCharArray(password_send,bu_password.length() + 1);
    WiFi.begin(ssid_send, password_send);
#ifdef DEBUG
    Serial.print("\r\nWiFi BU\r\n");
    Serial.println(ssid_send);
    Serial.println(password_send);
#endif
    for (int i=0; i<WIFI_TIMEOUT; i++)
    {
      if (WiFi.status() != WL_CONNECTED)
      {
#ifdef DEBUG
        Serial.print(".");
#endif
        delay(500);
      }
     if (WiFi.status() == WL_CONNECTED)
      {
        IPAddress myAddr = WiFi.localIP();
#ifdef DEBUG
        Serial.println("BU connected");
        Serial.println(myAddr);
#endif
      }
    }
  }  
  if (WiFi.status() != WL_CONNECTED)
  {  
    WiFi.mode(WIFI_AP);
    ap_name.toCharArray(ssid_send,ap_name.length() + 1);
    ap_password.toCharArray(password_send,ap_password.length() + 1);
    WiFi.softAP(ssid_send, password_send);
    IPAddress myAddr = WiFi.softAPIP();
#ifdef DEBUG
    Serial.println("\r\nAP-Mode Start");
    Serial.println(ssid_send);
    Serial.println(password_send);
    Serial.println(myAddr);
#endif
    ap_mode = true;
  }   
}


// ###############################################################
// ---------------------- Web-Socket events ----------------------
// ###############################################################

void webSocketEvent(uint8_t num, WStype_t type, uint8_t * payload, size_t length)
{
  float temperature;
  char buffer[20];
  String clock;
  
  switch(type) {
    case WStype_DISCONNECTED:
#ifdef DEBUG
      Serial.printf("[%u] Disconnected!\r\n", num);
#endif
      break;
    case WStype_CONNECTED:
      {
        IPAddress ip = webSocket.remoteIP(num);
#ifdef DEBUG
        Serial.printf("[%u] Connected from %d.%d.%d.%d url: %s\r\n", num, ip[0], ip[1], ip[2], ip[3], payload);
#endif
        // Send the current LED status
        switch (RelayStat) {
          case STEAM:
            webSocket.sendTXT(num, sSTEAM, strlen(sSTEAM));
          break;
          case HEAT:
            webSocket.sendTXT(num, sHEAT, strlen(sHEAT));
          break;
          case READY:
            webSocket.sendTXT(num, sREADY, strlen(sREADY));
          break;
          case OFF:
            webSocket.sendTXT(num, sOFF, strlen(sOFF));
          break;
          default:
#ifdef DEBUG
            Serial.println("Unknown RelayStat");
#endif
          break;
        }
      }    
      break;
    case WStype_TEXT:
      if (strncmp(sSTEAM, (const char *)payload, 3) == 0) {
        SetCoffee(STEAM, num);
      }
      else if (strncmp(sHEAT, (const char *)payload, 3) == 0) {
        SetCoffee(HEAT, num);
      }
      else if (strncmp(sOFF, (const char *)payload, 3) == 0) {
        SetCoffee(OFF, num);
      }
      else if (strncmp(sREAD, (const char *)payload, 3) == 0) {
        DS18B20.requestTemperatures(); 
        temperature = DS18B20.getTempCByIndex(0);
        dtostrf(temperature, 4, 1, buffer);
        for (int i = strlen(buffer)+2; i >= 0; i--){
          buffer[i] = buffer[i-2];
        }
        buffer[0]='T';
        buffer[1]='=';
        webSocket.sendTXT(num, buffer, strlen(buffer));
        sprintf (buffer, "%i", Coffeetimer);
        for (int i = strlen(buffer)+2; i >= 0; i--){
          buffer[i] = buffer[i-2];
        }
        buffer[0]='t';
        buffer[1]='=';
        webSocket.sendTXT(num, buffer, strlen(buffer));

//        clock = timeClient.getFormattedTime();
//        clock=timeClient.getFormattedTime() + " " + DayOfWeek[tdate->wday] + " " + tdate->year + "/" + tdate->month + "/" + tdate->day;

//      clock=timeClient.getFormattedTime() + " " + tdate->year + "/" + tdate->month + "/" + tdate->day;
//        clock.toCharArray(buffer, 20);
//        for (int i = strlen(buffer)+2; i >= 0; i--){
//          buffer[i] = buffer[i-2];
//        }
//        buffer[0]='c';
//        buffer[1]='=';
//        webSocket.sendTXT(num, buffer, strlen(buffer));

        switch (RelayStat) {
          case STEAM:
            webSocket.sendTXT(num, sSTEAM, strlen(sSTEAM));
          break;
          case HEAT:
            webSocket.sendTXT(num, sHEAT, strlen(sHEAT));
          break;
          case READY:
            webSocket.sendTXT(num, sREADY, strlen(sREADY));
          break;
          case OFF:
            webSocket.sendTXT(num, sOFF, strlen(sOFF));
          break;
          default:
#ifdef DEBUG
            Serial.println("Unknown RelayStat");
#endif
          break;
        }

      } else {
#ifdef DEBUG
        Serial.println("Unknown command");
#endif
      }
      // send data to all connected clients
      webSocket.broadcastTXT(payload, length);
      break;
    case WStype_BIN:
#ifdef DEBUG
      Serial.printf("[%u] get binary length: %u\r\n", num, length);
#endif
      hexdump(payload, length);

      // echo data back to browser
      webSocket.sendBIN(num, payload, length);
      break;
    default:
#ifdef DEBUG
      Serial.printf("Invalid WStype [%d]\r\n", type);
#endif
      break;
  }
}

void SetCoffee(uint8_t Cstat, uint8_t Snum)
{
#ifdef DEBUG
  Serial.print(Cstat);
#endif
  CoffeeStat = Cstat;
  if (CoffeeStat==OFF){
    webSocket.sendTXT(Snum, "t=0", 4);
    Coffeetimer=0;
  } else {
    Coffeetimer=CoffeeTimerValue;
  }
}


// ###############################################################
// ---------------------- Web-Tools ------------------------------
// ###############################################################

void handleUnknown()
{
  String filename = webserver.uri();

#ifdef DEBUG
  Serial.print("URL: ");
  Serial.println(filename);
#endif

  File pageFile = SPIFFS.open(filename, "r"); 
  if (pageFile)
  {
    String contentTyp = getContentType(filename);

#ifdef DEBUG
    Serial.print("Content-Typ: ");
    Serial.println(contentTyp);
#endif

    size_t sent = webserver.streamFile(pageFile, contentTyp);
    pageFile.close(); 
  } else {
    notFound(); 
  }
}

//---------------------------------------------------------------

void notFound()
{
  String sHTML = F("<html><head><title>404 Not Found</title></head><body><h1>Not Found</h1><p>The requested URL was not found on this webserver.</p></body></html>");
  webserver.send(404, "text/html", sHTML);    
}

// ----------------------------------------------------------------------

String GetPicture(String filename)
{   
  String contentTyp = getContentType(filename);

  File tempFile = SPIFFS.open(filename,"r");     
  if (tempFile)
  {
    if (tempFile.size() < MAX_PIC_SIZE)
    {        
      filename = tempFile.readString(); 
      filename = "data:" + contentTyp + ";base64," + base64::encode((uint8_t *) filename.c_str(), tempFile.size());  
#ifdef DEBUG
      Serial.println("Test");
      Serial.println(filename);
#endif
    }
    tempFile.close();
  }
  return filename;
}

// ----------------------------------------------------------------------

String getContentType(String filename)
{
  if(webserver.hasArg("download"))
    return "application/octet-stream";
  else if(filename.endsWith(".ico"))
    return "image/x-icon";
  else if(filename.endsWith(".htm"))
    return "text/html";
  else if(filename.endsWith(".html"))
    return "text/html";
  else if(filename.endsWith(".css"))
    return "text/css";
  else if(filename.endsWith(".js"))
    return "application/javascript";
  else if(filename.endsWith(".png"))
    return "image/png";
  else if(filename.endsWith(".gif"))
    return "image/gif";
  else if(filename.endsWith(".jpg"))
    return "image/jpeg";
  else if(filename.endsWith(".ico"))
    return "image/x-icon";
  else if(filename.endsWith(".xml"))
    return "text/xml";
  else if(filename.endsWith(".pdf"))
    return "application/x-pdf";
  else if(filename.endsWith(".zip"))
    return "application/x-zip";
  else if(filename.endsWith(".gz"))
    return "application/x-gzip";
  return "text/plain";
}

// ###############################################################
//---------------------- Webseiten -------------------------------
// ###############################################################

void handleUpload()
{

#ifdef DEBUG
  Serial.println("Datei-Upload"); 
#endif

  static File fsUploadFile;

  HTTPUpload& upload = webserver.upload();
  if(upload.status == UPLOAD_FILE_START)
  {
    String dateiname = upload.filename;
    if(!dateiname.startsWith("/"))
    {
      dateiname = "/" + dateiname;
    }
    
#ifdef DEBUG
    Serial.print("handleFileUpload Name: ");
    Serial.println(dateiname);
#endif

    fsUploadFile = SPIFFS.open(dateiname, "w");
    dateiname = String();
  }
  else if(upload.status == UPLOAD_FILE_WRITE)
  {
    if(fsUploadFile)
    {
      fsUploadFile.write(upload.buf, upload.currentSize);

#ifdef DEBUG
      Serial.println(upload.currentSize,DEC);
#endif

    }  
  }
  else if(upload.status == UPLOAD_FILE_END)
  {
    if(fsUploadFile)
    {
      fsUploadFile.close();
    }  

#ifdef DEBUG
    Serial.print("handleFileUpload Size: ");
    Serial.println(upload.totalSize);
#endif

  }
}

//---------------------------------------------------------------

void handleConfig()
{ 
  static File fsUploadFile;                                    // actual upload

  String filename = webserver.uri();

  File pageFile = SPIFFS.open(filename, "r"); 
  if (pageFile)
  {
    String sHTML = pageFile.readString();
    pageFile.close();

    String bilddaten = "/logo.jpg";
    bilddaten = GetPicture(bilddaten);
    sHTML.replace("~logo~",bilddaten);                         // insert pic on web page
        
    if (webserver.hasArg("reboot"))                           
    {
#ifdef DEBUG
      Serial.println("\r\n\r\n--- Neustart ---\r\n");
#endif
      ESP.restart();
    }
       
    // config.txt speichern
    File configfile;
    if (webserver.hasArg("config"))                            // new config.txt
    {

#ifdef DEBUG
      Serial.println("\r\n--- config.txt begin ---");
      Serial.print(webserver.arg("config"));  
      Serial.println("--- config.txt end ---\r\n");
#endif

      configfile = SPIFFS.open("/config.txt","w");             // overwrite config.txt
      if (configfile)
      {
        configfile.print(webserver.arg("config"));             // write data   
        configfile.close();   
        sHTML.replace("~config~",webserver.arg("config"));     // insert data in web page
      }
    } else {                                                   // write no config data
      File configfile = SPIFFS.open("/config.txt","r");        // get config.txt
      if (configfile)
      {     
        sHTML.replace("~config~",configfile.readString());     // insert data in web page
        configfile.close();
      } else {
        sHTML.replace("~config~","");                          // config.txt not exist
      }
    }

    // Datei download
    if (webserver.hasArg("filename"))                          // Download or delete
    {
      if (webserver.hasArg("download"))
      {
        File downfile = SPIFFS.open(webserver.arg("filename"),"r"); 
        String sendname = webserver.arg("filename");
        sendname.replace("/",""); 
        webserver.sendHeader("Content-Disposition"," attachment; filename=\"" + sendname + "\"");
        size_t sent = webserver.streamFile(downfile,"application/octet-stream");
        downfile.close();
#ifdef DEBUG
        Serial.print("Download Filename: ");
        Serial.println(sendname);
#endif
      }
      if (webserver.hasArg("delete"))
      {
#ifdef DEBUG
        Serial.print("Delete Filename: ");
#endif
        SPIFFS.remove(webserver.arg("filename"));
      }
#ifdef DEBUG
      Serial.println(webserver.arg("filename"));
#endif
    }
    
#ifdef DEBUG
    Serial.println("\r\n--- Dir-Liste begin ---");
#endif

    // Dateiliste anzeigen
    Dir dir = SPIFFS.openDir("/");
    int anzahl = 0;
    while (dir.next())
    {
      String filename = dir.fileName(); 
      File dirFile = dir.openFile("r");
      size_t filelen =  dirFile.size(); 

#ifdef DEBUG
      Serial.print(filename);
      Serial.print(" - ");
      Serial.println(filelen);
#endif
      String filename_temp = filename;
      filename_temp.replace("/","");
      String dirzeile = "<form action=\"config.html\" method=\"POST\">\r\n<tr><td align=\"center\">";
      dirzeile += "<input type=\"hidden\" name = \"filename\" value=\"" + filename + "\">" + filename_temp;
      dirzeile +=  "</td>\r\n<td align=\"right\">" + String(filelen,DEC);
      dirzeile += "</td><td align=\"center\"><button type=\"submit\" name= \"download\">Download</button></td>\r\n";
      dirzeile += "<td align=\"center\" bgcolor=\"red\"><button type=\"submit\" name=\"delete\">&nbsp;&nbsp;&nbsp;Delete&nbsp;&nbsp;&nbsp;</button></td>";
      dirzeile += "</tr></form>\r\n~dirzeile~"; 
      sHTML.replace("~dirzeile~",dirzeile);
      anzahl++;
    }   
    sHTML.replace("~dirzeile~","");

    FSInfo fs_info;
    SPIFFS.info(fs_info);
    sHTML.replace("~anzahl~",String(anzahl,DEC));
    sHTML.replace("~total~",String(fs_info.totalBytes/1024,DEC));
    sHTML.replace("~used~",String(fs_info.usedBytes/1024,DEC));
    sHTML.replace("~free~",String((fs_info.totalBytes - fs_info.usedBytes)/1024,DEC));

#ifdef DEBUG
  Serial.println("--- Dir-Liste end ---\r\n");
#endif
   
    webserver.send(200, "text/html", sHTML);
  } else {
    notFound(); 
  }
}


//---------------------------------------------------------------


void handleCoffee()
{
  String filename = webserver.uri();
    if (filename == "/")
  {
    filename = "/coffee.html";
  }
  
  File pageFile = SPIFFS.open(filename,"r"); 
  if (pageFile)
  {
    String sHTML = pageFile.readString();
    pageFile.close();  

    webserver.send(200, "text/html", sHTML);
  } else {
    notFound(); 
  }
}

//---------------------------------------------------------------
// https://www.mikrocontroller.net/topic/258193
//void gettime( uint32_t sec, struct time *t )
//{
//    uint16_t day;
//    uint8_t  year;
//    uint16_t dayofyear;
//    uint8_t  leap400;
//    uint8_t  month;
//
//    t->second = sec % 60;
//    sec      /= 60;
//    t->minute = sec % 60;
//    sec      /= 60;
//    t->hour   = sec % 24;
//    day       = sec / 24;
//    t->wday = (day + FIRSTDAY) % 7;             // weekday
//
//    year = FIRSTYEAR % 100;                     // 0..99
//    leap400 = 4 - ((FIRSTYEAR - 1) / 100 & 3);  // 4, 3, 2, 1
//
//    for(;;)
//    {
//        dayofyear = 365;
//        if( (year & 3) == 0 )
//        {
//            dayofyear = 366;                    // leap year
//            if( year == 0 || year == 100 || year == 200 ) // 100 year exception
//                if( --leap400 )                 // 400 year exception
//                    dayofyear = 365;
//        }
//        if( day < dayofyear )
//            break;
//        day -= dayofyear;
//        year++;                                 // 00..136 / 99..235
//    }
//    t->year = year + (int)(FIRSTYEAR / 100) * 100;     // + century
//
//    if( dayofyear & 1 && day > 58 )             // no leap year and after 28.2.
//        day++;                                  // skip 29.2.
//
//    for( month = 1; day >= DayOfMonth[month-1]; month++ )
//        day -= DayOfMonth[month-1];
//
//    t->month = month;                           // 1..12
//    t->day   = day + 1;                         // 1..31
//}

